import numpy
import pytest
from signalworks.tracking.timevalue import TimeValue
from signalworks.tracking.tracking import Track

TIME_TYPE = numpy.int64


@pytest.fixture
def var():
    t1 = TimeValue(
        (numpy.linspace(1, 9, 3)).astype(TIME_TYPE), numpy.array([1, 4, 2]), 1, 10
    )
    t2 = TimeValue(
        (numpy.linspace(2, 8, 4)).astype(TIME_TYPE), numpy.array([1, 4, 8, 2]), 1, 10
    )
    assert isinstance(t1, Track)
    assert isinstance(t2, Track)
    return t1, t2


def test_init():
    with pytest.raises(Exception):
        TimeValue(
            numpy.array([6, 3], dtype=TIME_TYPE), numpy.array([3, 6]), 1, 10
        )  # bad times
    with pytest.raises(Exception):
        TimeValue(
            numpy.array([3, 6], dtype=TIME_TYPE), numpy.array([3, 6]), 1, 5
        )  # duration too short


def test_eq(var):
    t1, t2, = var
    assert t1 == t1
    assert not t1 == t2


def test_add(var):
    t1, t2, = var
    t = t1 + t2
    assert t.duration == 20
    assert t.time[5] == 16
    assert t.value[5] == 8
    t1 += t2
    t = t1
    assert t.duration == 20
    assert t.time[5] == 16
    assert t.value[5] == 8


def test_select(var):
    t1, _ = var
    t = t1.select(1, 5)
    assert t == TimeValue(numpy.array([0], dtype=TIME_TYPE), numpy.array([1]), 1, 4)
    t = t1.select(1, 6)
    assert t == TimeValue(
        numpy.array([0, 4], dtype=TIME_TYPE), numpy.array([1, 4]), 1, 5
    )
    t = t1.select(1, 6)
    assert t == TimeValue(
        numpy.array([0, 4], dtype=TIME_TYPE), numpy.array([1, 4]), 1, 5
    )
    # t = t1.select(2, 5)
    # assert t == TimeValue(numpy.array([], dtype=TIME_TYPE), numpy.array([]), 1, 3) # empty value


def test_duration(var):
    t1, _ = var
    t1.duration = 11  # ok
    with pytest.raises(Exception):
        t1.set_duration(5)  # duration too short



import numpy as np
import pytest
from signalworks.tracking import Label, Track


@pytest.fixture
def var():
    p1 = Label(
        np.array([0, 3, 5, 8, 10, 12], dtype=np.int64),
        np.array(["start", "middle", "end"]),
        1,
        12,
    )
    p2 = Label(
        np.array([0, 5, 8, 10], dtype=np.int64), np.array(["start2", "end2"]), 1, 10
    )
    assert isinstance(p1, Track)
    assert isinstance(p2, Track)
    return p1, p2


def test_eq(var):
    p1, p2 = var
    assert p1 == p1
    assert not p1 == p2


def test_add(var):
    p1, p2 = var
    p = p1 + p2
    assert p.duration == 22
    assert p.time[7] == 17
    assert p.value[4] == "end2"
    p1 += p2
    p = p1
    assert p.duration == 22
    assert p.time[7] == 17
    assert p.value[4] == "end2"



# -*- coding: utf-8 -*-
import distutils.spawn
from pathlib import Path

import pytest  # noqa: F401

from signalworks.tracking import Track


def has_gitlfs() -> bool:
    return distutils.spawn.find_executable("git-lfs") is not None


def is_gitlfs_pointer(path: Path) -> bool:
    return path.stat().st_blocks == 8 and path.stat().st_blksize == 4096


xfailif_no_gitlfs = pytest.mark.xfail(
    not has_gitlfs(), reason="This test requires git-lfs"
)


@xfailif_no_gitlfs
@pytest.fixture(scope="function")
def speech_track():
    speech_wav = Path(__file__).parents[1] / "data" / "speech-mwm.wav"
    if is_gitlfs_pointer(speech_wav):
        return None
    else:
        track = Track.read(speech_wav)
    assert isinstance(track, Track)
    return track



import numpy as np
import pytest
from signalworks.tracking import Partition, TimeValue, Track


@pytest.fixture
def var():
    p1 = Partition(
        np.array([0, 5, 6, 10], dtype=np.int64), np.array(["start", "middle", "end"]), 1
    )
    p2 = Partition(
        np.array([0, 5, 10], dtype=np.int64), np.array(["start2", "end2"]), 1
    )
    assert isinstance(p1, Track)
    assert isinstance(p2, Track)
    return p1, p2


def test_insert():
    p = Partition(
        np.array([0, 5, 6, 10], dtype=np.int64), np.array(["start", "middle", "end"]), 1
    )
    p.insert(7, "still the middle")
    assert (p.time == np.array([0, 5, 6, 7, 10])).all()


def test_eq(var):
    p1, p2 = var
    assert p1 == p1
    assert not p1 == p2


def test_add(var):
    p1, p2 = var
    p = p1 + p2
    assert p.duration == 20
    assert p.time[4] == 15
    assert p.value[4] == "end2"
    p1 += p2
    p = p1
    assert p.duration == 20
    assert p.time[4] == 15
    assert p.value[4] == "end2"


def test_select(var):
    p1, p2 = var
    p = p1.select(5, 6)
    assert p == Partition(np.array([0, 1], dtype=np.int64), np.array(["middle"]), 1)
    p = p1.select(5, 7)
    assert p == Partition(
        np.array([0, 1, 2], dtype=np.int64), np.array(["middle", "end"]), 1
    )
    p = p1.select(4, 6)
    assert p == Partition(
        np.array([0, 1, 2], dtype=np.int64), np.array(["start", "middle"]), 1
    )
    p = p1.select(4, 7)
    assert p == Partition(
        np.array([0, 1, 2, 3], dtype=np.int64), np.array(["start", "middle", "end"]), 1
    )


def test_from_TimeValue(benchmark):
    tv = TimeValue(
        np.arange(9, dtype=np.int64) * 10 + 10,
        np.array([0.0, 1.0, 1.0, 4.0, 4.0, 4.0, 8.0, 8.0, 8.0]),
        1,
        100,
    )
    p = benchmark(Partition.from_TimeValue, tv)
    assert (p.time == np.array([0, 15, 35, 65, 100])).all()
    assert (p.value == np.array([0.0, 1.0, 4.0, 8.0])).all()


def test_merge_same(benchmark):
    p = Partition(np.array([0, 3, 6, 10], dtype=np.int64), np.array(["1", "1", "2"]), 1)
    p = benchmark(p.merge_same)
    assert p.value[1] == "2"


def test_crossfade(benchmark):
    prt1 = Partition(np.array([0, 8, 10], dtype=np.int64), np.array(["1", "2"]), 1)
    prt2 = Partition(np.array([0, 2, 10], dtype=np.int64), np.array(["3", "4"]), 1)
    length = 4
    prt = benchmark(prt1.crossfade, prt2, length)
    assert prt1.duration + prt2.duration - length == prt.duration
    assert np.allclose(prt.time, np.array([0, 8, 16]))
    assert (prt.value == np.array(["1", "4"])).all()



# -*- coding: utf-8 -*-
from typing import NamedTuple, Optional, Tuple

import numpy as np
from signalworks import dsp
from signalworks.processors.processing import DefaultProgressTracker, Processor
from signalworks.tracking import TimeValue, Wave


class SpectralDiscontinuityEstimator(Processor):
    name = "Spectral Discontinuity Estimator"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def __init__(self):
        super().__init__()
        self.parameters = {
            "frame_size": 0.005,  # seconds, determines freq res.
            "NFFT": 256,
            "normalized": 1,
            "delta_order": 1,
        }

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[TimeValue]:
        if progressTracker is not None:
            self.progressTracker = progressTracker
        wav = self.data.wave
        assert isinstance(wav, Wave)
        self.progressTracker.update(10)
        ftr, time, frequency = dsp.spectrogram(
            wav,
            self.parameters["frame_size"],
            self.parameters["frame_size"],  # frame_rate = frame_size
            NFFT=self.parameters["NFFT"],
            normalized=self.parameters["normalized"],
        )
        if self.parameters["normalized"]:
            ftr = ftr - np.mean(ftr, axis=1).reshape(-1, 1)

        time = (time[:-1] + time[1:]) // 2
        assert self.parameters["delta_order"] > 0
        dynamic_win = np.arange(
            -self.parameters["delta_order"], self.parameters["delta_order"] + 1
        )

        win_width = self.parameters["delta_order"]
        win_length = 2 * win_width + 1
        den = 0
        for s in range(1, win_width + 1):
            den += s ** 2
        den *= 2
        dynamic_win = dynamic_win / den

        N, D = ftr.shape
        print(N)
        temp_array = np.zeros((N + 2 * win_width, D))
        delta_array = np.zeros((N, D))
        self.progressTracker.update(90)
        temp_array[win_width : N + win_width] = ftr
        for w in range(win_width):
            temp_array[w, :] = ftr[0, :]
            temp_array[N + win_width + w, :] = ftr[-1, :]

        for i in range(N):
            for w in range(win_length):
                delta_array[i, :] += temp_array[i + w, :] * dynamic_win[w]
        value = np.mean(np.diff(delta_array, axis=0) ** 2, axis=1) ** 0.5
        dis = TimeValue(
            time,
            value,
            wav.fs,
            wav.duration,
            path=wav.path.with_name(wav.path.stem + "-discont").with_suffix(
                TimeValue.default_suffix
            ),
        )
        dis.min = 0
        dis.max = value.max()
        dis.unit = "dB"
        dis.label = "spectral discontinuity"
        self.progressTracker.update(100)
        return (dis,)



# -*- coding: utf-8 -*-
"""
Digital Signal Processing
"""
from math import ceil, log2
from typing import Tuple, Union

import numba
import numpy as np
from numpy.fft import fft, ifft, irfft, rfft
from scipy import signal, stats
from signalworks.tracking import TimeValue, Wave

# TODO: make this "tracking"-free (?), and all times are in samples

# def segment_talkbox(a: np.ndarray,
#                     length: int,
#                     overlap: int = 0) -> np.ndarray:
#     # originally from talkbox.segmentaxis
#     a = np.ravel(a)  # may copy
#     l = a.shape[0]
#     if overlap >= length:
#         raise ValueError("frames cannot overlap by more than 100%")
#     if overlap < 0 or length <= 0:
#         raise ValueError("overlap must be nonnegative and length must be positive")
#     if l < length or (l - length) % (length - overlap):
#         if l > length:
#             roundup = length + (1 + (l - length) //
#                                 (length - overlap)) * (length - overlap)
#             # TODO: further optimization possible
#             rounddown = length + ((l - length) // (length - overlap)) * (length - overlap)
#         else:
#             roundup = length
#             rounddown = 0
#         assert rounddown < l < roundup
#         assert roundup == rounddown + (length - overlap) or (roundup == length and rounddown == 0)
#         a = a.swapaxes(-1, 0)
#         a = a[..., :rounddown]
#         a = a.swapaxes(-1, 0)
#     l = a.shape[0]
#     if l == 0:
#         raise ValueError("Not enough data points to segment array")
#     assert l >= length
#     assert (l - length) % (length - overlap) == 0
#     n = 1 + (l - length) // (length - overlap)
#     s = a.strides[0]
#     newshape = a.shape[:0] + (n, length) + a.shape[1:]
#     newstrides = a.strides[:0] + ((length - overlap) * s, s) + a.strides[1:]
#     try:
#         return np.ndarray.__new__(np.ndarray, strides=newstrides,
#                                   shape=newshape, buffer=a, dtype=a.dtype)
#     except TypeError:
#         import warnings
#         warnings.warn("Problem with ndarray creation forces copy.")
#         a = a.copy()
#         # Shape doesn't change but strides does
#         newstrides = a.strides[:0] + ((length - overlap) * s, s) + a.strides[1:]
#         return np.ndarray.__new__(np.ndarray,
#                                   strides=newstrides,
#                                   shape=newshape,
#                                   buffer=a,
#                                   dtype=a.dtype)


# @numba.jit((numba.int16[:], numba.int64, numba.int32), nopython=True, cache=True)
@numba.jit(nopython=True, cache=True)  # we need polymorphism here
def segment(x, nsize, nrate):
    if len(x) < nsize:
        F = 0
    else:
        F = (len(x) - nsize) // nrate  # the number of full frames
    assert F >= 0
    X = np.empty((F, nsize), dtype=x.dtype)
    a = 0
    for f in range(F):
        X[f, :] = x[a : a + nsize]
        a += nrate
    return X


def frame(wav: Wave, frame_size: float, frame_rate: float) -> TimeValue:
    """
    Given a waveform, return a timeValue track with each frame as the value and times of the center of each frame.
    times point to the center of the frame.
    Each frame will have the specified size, and t[i+1] = t[i] + rate.
    this will return as much of the signal as possible in full frames
    """
    # def unsigned int a, f, nrate, nsize
    assert wav.duration > 0
    nsize = int(round(frame_size * wav.fs))
    nrate = int(round(frame_rate * wav.fs))
    # import time
    # tic = time.time()
    # print("frame timing...")
    # if 0:  # TODO: unfortunately segment doesn't allow for negative overlap, i.e. jumps
    #     value = segment_talkbox(wav.value, nsize, nsize - nrate)  # because overlap = nsize - nrate
    value = segment(wav.value, nsize, nrate)
    # print(f"frame took time: {time.time() - tic}")
    assert value.shape[1] == nsize
    time = np.array(np.arange(value.shape[0]) * nrate, dtype=np.int64) + nsize // 2
    return TimeValue(
        time, value, wav.fs, wav.duration, path=wav.path
    )  # adjust path name here?


# @numba.jit(nopython=True, cache=True)  # we need polymorphism here
def frame_centered(signal: np.ndarray, time: np.ndarray, frame_size: int) -> np.ndarray:
    assert time.ndim == 1
    # no further assumptions on time - doesn't have to be sorted or inside signal
    value = np.zeros((len(time), frame_size), dtype=signal.dtype)
    left_frame_size = frame_size // 2
    right_frame_size = frame_size - left_frame_size
    S = len(signal)
    for f, center in enumerate(time):
        left = center - left_frame_size
        right = center + right_frame_size
        if left >= 0 and right <= S:  # make the common case fast
            value[f, :] = signal[left:right]
        else:  # deal with edges on possibly both sides
            # left
            if left < 0:
                left_avail = left_frame_size + left
            else:
                left_avail = left_frame_size
            # right
            right_over = right - S
            if right_over > 0:
                right_avail = right_frame_size - right_over
            else:
                right_avail = right_frame_size
            if 0 <= center <= S:
                value[
                    f, left_frame_size - left_avail : left_frame_size + right_avail
                ] = signal[center - left_avail : center + right_avail]
    assert value.shape[0] == len(time)
    assert value.shape[1] == frame_size
    return value  # adjust path name here?


@numba.jit(nopython=True, cache=True)  # we need polymorphism here
def ola(
    frame: np.ndarray, fs: int, duration: int, frame_size: float, frame_rate: float
) -> np.ndarray:
    nsize = int(round(frame_size * fs))
    nrate = int(round(frame_rate * fs))
    y = np.zeros(duration, dtype=np.float64)
    a = 0
    for f in range(len(frame)):
        y[a : a + nsize] += frame[f]
        a += nrate
    return y


def spectral_subtract(inp: Wave, frame_rate: int, silence_percentage: int) -> Wave:
    assert 0 < silence_percentage < 100
    ftr = frame(inp, frame_rate * 2, frame_rate)
    x = ftr.value * signal.hann(ftr.value.shape[1])
    X = fft(x, 2 ** nextpow2(x.shape[1]))
    M = np.abs(X)
    E = np.mean(M ** 2, axis=1)
    threshold = stats.scoreatpercentile(E, silence_percentage)
    index = np.where(E < threshold)[0]
    noise_profile = np.median(M[index], axis=0)
    M -= noise_profile
    np.clip(
        M, 0, None, out=M
    )  # limit this to a value greater than 0 to avoid -inf due to the following log
    Y = M * np.exp(1j * np.angle(X))  # DEBUG
    y = ifft(Y).real
    s = ola(y, inp.fs, inp.duration, frame_rate * 2, frame_rate)
    return Wave(s, inp.fs)


def spectrogram(
    wav: Wave,
    frame_size: float,
    frame_rate: float,
    window: signal.hann = signal.hann,
    NFFT: str = "nextpow2",
    normalized: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """return log-magnitude spectrogram in dB"""
    ftr = frame(wav, frame_size, frame_rate)
    x = ftr.value * window(ftr.value.shape[1])
    if NFFT == "nextpow2":
        NFFT = 2 ** nextpow2(x.shape[1])
    M = np.abs(rfft(x, NFFT))
    np.clip(M, 1e-12, None, out=M)
    M = np.log10(M) * 20
    if normalized:
        M = (M.T - np.min(M, axis=1)).T
        M = (M.T / np.max(M, axis=1)).T
        assert np.all(M.min(axis=1) == 0)
        assert np.all(M.max(axis=1) == 1)
    frequency = np.arange(M.shape[1]) / M.shape[1] * wav.fs / 2
    return M, ftr.time, frequency


def spectrogram_centered(
    wav: Wave,  # used by rendering
    frame_size: float,
    time: np.ndarray,
    window: signal.hann = signal.hann,
    NFFT: str = "nextpow2",
    normalized: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """return log-magnitude spectrogram in dB"""
    s = wav.value / np.abs(
        np.max(wav.value)
    )  # make float by normalizing and later clipping is more uniform
    assert s.max() == 1
    ftr = frame_centered(s, time, int(round(frame_size * wav.fs)))
    assert ftr.dtype == np.float
    ftr *= window(ftr.shape[1])
    if NFFT == "nextpow2":
        NFFT = 2 ** nextpow2(ftr.shape[1])
    M = np.abs(rfft(ftr, NFFT))
    np.clip(M, 1e-16, None, out=M)
    M[:] = np.log10(M) * 20
    if normalized:
        M[:] = (M.T - np.min(M, axis=1)).T
        M[:] = (M.T / np.max(M, axis=1)).T
        # assert np.all(M.min(axis=1) == 0)
        # assert np.all(M.max(axis=1) == 1)
    frequency = np.arange(M.shape[1]) / M.shape[1] * wav.fs / 2
    return M, frequency


def correlate_fft(X: np.ndarray) -> np.ndarray:
    """correlation for feature matrix"""
    assert X.ndim == 2
    D = X.shape[1]
    R = irfft(np.abs(rfft(X, 2 ** nextpow2(2 * D - 1))) ** 2)[:, :D]
    # show relationship to related methods
    assert np.allclose(R[0], np.correlate(X[0], X[0], mode="full")[D - 1 :])
    # assert np.allclose(r, np.convolve(x, x[::-1], mode='full')[n - 1:])
    # from scipy.signal import fftconvolve
    # assert np.allclose(r, fftconvolve(x, x[::-1], mode='full')[n - 1:])
    return R


def correlogram(
    wav: Wave, frame_size: float, frame_rate: float, normalize: bool = True
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    assert wav.dtype == np.float64
    # t, x = frame(wav, frame_size, frame_rate)
    ftr = frame(wav, frame_size, frame_rate)
    M, D = ftr.value.shape
    R = correlate_fft(ftr.value)
    # if 1:
    #     # FFT order must be at least 2*len(x)-1 and should be a power of 2
    #     R = irfft(np.abs(rfft(ftr.value, 2 ** nextpow2(2 * D - 1))) ** 2)[:, D]
    # else:
    #     index = np.arange(int(np.round(D / 2)), D)
    #     R = np.empty((M, len(index)), dtype=np.float64)
    #     for m in range(M):
    #         signal = ftr.value[m]
    #         R[m, :] = np.correlate(signal, signal, mode='same')[index]  # TODO: use fft2 here instead
    if normalize:
        R[:, 1:] /= np.tile(
            R[:, 0], (R.shape[1] - 1, 1)
        ).T  # keep energy in zero-th coeff?
    frequency = np.r_[np.nan, wav.fs / np.arange(1, R.shape[1])]
    return R, ftr.time, frequency


def nextpow2(i: Union[int, float]) -> int:
    """returns the first P such that 2**P >= abs(N)"""
    return int(ceil(log2(i)))


#
#
# def world(wave: Wave) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
#     x = wave.value / (2 ** 15)
#     fs = wave.fs
#     _f0, t = pw.dio(x, fs)  # raw pitch extractor
#     f0 = pw.stonemask(x, _f0, t, fs)  # pitch refinement
#     sp = pw.cheaptrick(x, f0, t, fs)  # extract smoothed spectrogram
#     return sp, f0, t



import abc


class MetaTrack(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_duration(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-
import numpy as np
import pytest

from signalworks.processors import (
    ConverterToFloat64,
    EnergyEstimator,
    F0Analyzer,
    Filter,
    NoiseReducer,
    PeakTracker,
    ZeroPhaseFilter,
)
from signalworks.tracking import TimeValue, Track, Wave


def test_filter(speech_track, benchmark):
    if speech_track is None:
        pytest.skip("the test need wavefile")
    processor = Filter()
    processor.set_data({"wave": speech_track})
    results, = benchmark(processor.process)
    assert speech_track.duration == results.duration
    assert isinstance(results.get_value(), np.ndarray)
    assert isinstance(results, Wave)
    assert isinstance(results, Track)


def test_zerofilter(speech_track, benchmark):
    if speech_track is None:
        pytest.skip("the test need wavefile")
    zp_processor = ZeroPhaseFilter()
    zp_processor.set_data({"wave": speech_track})
    results, = benchmark(zp_processor.process)
    assert speech_track.duration == results.duration
    assert isinstance(results.get_value(), np.ndarray)
    assert isinstance(results, Wave)
    assert isinstance(results, Track)


def test_f0(speech_track, benchmark):
    if speech_track is None:
        pytest.skip("the test need wavefile")
    processor = F0Analyzer()
    processor.set_data({"wave": speech_track})
    f0, dop, vox = benchmark(processor.process)
    assert f0.duration == dop.duration
    assert isinstance(f0, TimeValue)


def test_convert(speech_track, benchmark):
    if speech_track is None:
        pytest.skip("the test need wavefile")
    processor = ConverterToFloat64()
    processor.set_data({"wave": speech_track})
    results, = benchmark(processor.process)
    assert speech_track.duration == results.duration
    assert isinstance(results.get_value(), np.ndarray)


def test_energy(speech_track, benchmark):
    if speech_track is None:
        pytest.skip("the test need wavefile")
    processor = EnergyEstimator()
    processor.set_data({"wave": speech_track})
    results, = benchmark(processor.process)
    assert len(results.get_time()) == len(results.get_value())
    assert isinstance(results, TimeValue)


def test_peaktrack(speech_track, benchmark):
    if speech_track is None:
        pytest.skip("the test need wavefile")
    processor = PeakTracker()
    processor.set_data({"wave": speech_track})
    results, = benchmark(processor.process)
    assert len(results.get_time()) == len(results.get_value())
    assert isinstance(results, TimeValue)


def test_noisereducer(speech_track, benchmark):
    if speech_track is None:
        pytest.skip("the test need wavefile")
    processor = NoiseReducer()
    processor.set_data({"wave": speech_track})
    results, = benchmark(processor.process)
    assert len(results.get_time()) == len(results.get_value())
    assert isinstance(results, Wave)



# -*- coding: utf-8 -*-
import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Dict, NamedTuple, Optional, Tuple, Union

import numpy as np
from signalworks.tracking import Partition, TimeValue, Wave

Tracks = Union[Wave, TimeValue, Partition]
# type alias


class Data(NamedTuple):
    wave: Optional[Wave]
    active: Optional[Partition]


class InvalidDataError(Exception):
    pass


class InvalidParameterError(Exception):
    pass


class ProcessError(Exception):
    pass


class DefaultProgressTracker:
    def update(self, value: int) -> None:
        print(f"{value}%", end="...", flush=True)


class Processor(metaclass=ABCMeta):
    name = "Processor"
    acquire = NamedTuple("acquire", [("wave", Wave), ("active", Partition)])

    def __init__(self):
        self.data = Data(wave=None, active=None)
        self.parameters: Dict[str, Any] = {}  # default parameters
        self.progressTracker = DefaultProgressTracker()

    def set_data(self, data: Dict[str, Tracks]) -> None:
        wav = data.get("wave")
        if wav is not None and not isinstance(wav, Wave):
            raise InvalidDataError

        active = data.get("active")
        if active is not None and not isinstance(wav, Partition):
            raise InvalidDataError

        self.data = Data(wave=wav, active=active)

    def get_parameters(self) -> Dict[str, str]:
        # default parameters can be modified here based on the data
        return {k: str(v) for k, v in self.parameters.items()}

    def set_parameters(self, parameters: Dict[str, str]) -> None:
        if __debug__:
            for name, value in parameters.items():
                logging.debug(f"Received parameter {name} of value {value}")
        try:
            for key in parameters.keys():
                if type(self.parameters[key]) == np.ndarray:
                    self.parameters[key] = np.fromstring(
                        parameters[key].rstrip(")]").lstrip("[("), sep=" "
                    )
                else:
                    self.parameters[key] = type(self.parameters[key])(parameters[key])
        except Exception as e:
            raise InvalidParameterError(e)
        # additional parameter checking can be performed here

    @abstractmethod
    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[Tracks, ...]:
        pass

    def del_data(self):
        self.data = Data(wave=None, active=None)


def get_processor_classes() -> Dict[str, Callable[..., Processor]]:
    def all_subclasses(c):
        return c.__subclasses__() + [
            a for b in c.__subclasses__() for a in all_subclasses(b)
        ]

    return {obj.name: obj for obj in all_subclasses(Processor)}



import os
from pathlib import Path

import numpy
from signalworks.tracking.tracking import Track

TIME_TYPE = numpy.int64


class Value(Track):
    """can store a singular value (e.g. comment string) of any type"""

    default_suffix = ".trk"

    def __init__(self, value, fs, duration, path=None):
        super().__init__(path)
        if path is None:
            path = str(id(self))
        self.path = Path(path).with_suffix(self.default_suffix)
        assert isinstance(fs, int)
        assert fs > 0
        assert isinstance(duration, TIME_TYPE) or isinstance(duration, int)
        self._value = value
        self._fs = fs
        self._duration = duration

    def __getitem__(self, k):
        return self.value[k]

    def __setitem__(self, k, v):
        self.value[k] = v

    def __eq__(self, other):
        if (
            (self._fs == other._fs)
            and (self._duration == other._duration)
            and isinstance(other.value, type(self.value))
        ):
            return self.value == other.value
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return 1

    def __str__(self):
        return self._value

    def get_duration(self):
        return self._duration

    def set_duration(self, duration):
        assert isinstance(duration, TIME_TYPE) or isinstance(duration, int)
        self._duration = duration

    duration = property(get_duration, set_duration, doc="duration of track")

    def get_fs(self):
        return self._fs

    def set_fs(self, _value):
        raise Exception("Cannot change fs, try resample()")

    fs = property(get_fs, set_fs, doc="sampling frequency")

    def get_value(self):
        return self._value

    def set_value(self, value):
        self._value = value

    value = property(get_value, set_value)

    def select(self, a, b):
        raise NotImplementedError

    def time_warp(self, _x, y):
        self.duration = y[-1]  # changed this from _duration

    @classmethod
    def identify_ext(cls, name, track_name=None):
        if name == os.path.splitext(name)[0]:
            if track_name is None:
                raise ValueError("neither track name nor extension specified")
            elif track_name == "input":
                ext = ".txt"
            elif track_name == "xml":
                ext = ".xml"
            elif track_name == "search":
                ext = ".search"
            elif track_name == "pronunciation":
                ext = ".pron"
            else:
                raise ValueError("unknown track name '{}'".format(track_name))
        else:
            ext = os.path.splitext(name)[1]
        return ext

    @classmethod
    def read(cls, name, fs, duration, track_name=None, *_args, **_kwargs):
        """Loads object from name, adding extension if missing."""
        ext = cls.identify_ext(name, track_name)
        name = os.path.splitext(name)[0] + ext
        if ext in (".txt", ".xml", ".search"):
            with open(name, "r") as f:
                value = f.read()
                if ext == ".search":
                    value = str(value)  # , "UTF-8")
                self = Value(value, fs, duration)
        elif ext == ".pron":
            with open(name, "r") as f:
                pron = [tuple(line.split("\t")) for line in f.read().split("\n")]
                self = Value(pron, fs, duration)
        else:
            raise ValueError("unknown extension '{}'".format(ext))
        return self

    def write(self, name, track_name=None, *_args, **_kwargs):
        """
        Saves object to name, adding extension if missing.
        """
        ext = self.identify_ext(name, track_name)
        name = os.path.splitext(name)[0] + ext
        if ext in (".txt", ".xml", ".search"):
            with open(name, "w") as f:
                f.write(self.get_value())
        elif ext == ".pron":
            with open(name, "w") as f:
                f.write(
                    "\n".join(word + "\t" + pron for word, pron in self.get_value())
                )
        else:
            raise ValueError("unknown extension '{}'".format(ext))



# -*- coding: utf-8 -*-
import distutils.spawn
from pathlib import Path

import numpy as np
import pytest

from signalworks.tracking import load_audio


def has_gitlfs() -> bool:
    return distutils.spawn.find_executable("git-lfs") is not None


def is_gitlfs_pointer(path: Path) -> bool:
    return path.stat().st_blocks == 8 and path.stat().st_blksize == 4096


xfailif_no_gitlfs = pytest.mark.xfail(
    not has_gitlfs(), reason="This test requires git-lfs"
)


@xfailif_no_gitlfs
def test_load_wav():
    # read regular wav file
    path = Path(__file__).parents[1] / "data" / "speech-mwm.wav"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    multiTrack = load_audio(path)
    wave = multiTrack["left"]
    assert np.any(wave.value > 0)
    assert wave.fs == 22050


@xfailif_no_gitlfs
def test_load_au():
    # read au file
    soundfile = pytest.importorskip(  # noqa
        "soundfile", reason="If soundfile is not installed, this test will fail"
    )
    path = Path(__file__).parents[1] / "data" / "test.au"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    multiTrack = load_audio(path)
    wave = multiTrack["left"]
    assert np.any(wave.value > 0)
    assert wave.fs == 44100


@xfailif_no_gitlfs
def test_load_TIMIT():
    # read NIST file
    soundfile = pytest.importorskip(  # noqa
        "soundfile", reason="If soundfile is not installed, this test will fail"
    )
    path = Path(__file__).parents[1] / "data" / "test.WAV"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    multiTrack = load_audio(path)
    wave = multiTrack["left"]
    assert np.any(wave.value > 0)
    assert wave.fs == 16000


@xfailif_no_gitlfs
def test_load_nis():
    # read NIST file
    soundfile = pytest.importorskip(  # noqa
        "soundfile", reason="If soundfile is not installed, this test will fail"
    )
    path = Path(__file__).parents[1] / "data" / "test.nis"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    multiTrack = load_audio(path)
    wave = multiTrack["left"]
    assert np.any(wave.value > 0)
    assert wave.fs == 16000


@xfailif_no_gitlfs
def test_2channels():
    soundfile = pytest.importorskip(  # noqa
        "soundfile", reason="If soundfile is not installed, this test will fail"
    )
    path = Path(__file__).parents[1] / "data" / "test.flac"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    multiTrack = load_audio(path)
    wave_left = multiTrack["left"]
    assert np.any(wave_left.value > 0)
    assert wave_left.fs == 44100

    wave_right = multiTrack["right"]
    assert np.any(wave_right.value > 0)
    assert wave_right.fs == 44100


@xfailif_no_gitlfs
def test_load_wa1():
    # read WA1 file
    soundfile = pytest.importorskip(  # noqa
        "soundfile", reason="If soundfile is not installed, this test will fail"
    )
    path = Path(__file__).parents[1] / "data" / "test.wa1"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    multiTrack = load_audio(path)
    wave = multiTrack["left"]
    assert np.any(wave.value > 0)
    assert wave.fs == 8000


@xfailif_no_gitlfs
def test_load_wa2():
    # read WA2 file
    path = Path(__file__).parents[1] / "data" / "test.wa2"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    soundfile = pytest.importorskip(  # noqa
        "soundfile", reason="If soundfile is not installed, this test will fail"
    )
    multiTrack = load_audio(path)
    wave = multiTrack["left"]
    assert np.any(wave.value > 0)
    assert wave.fs == 8000


@xfailif_no_gitlfs
@pytest.mark.xfail(reason="We cannot support this kind of file")
def test_load_wv1():
    # read WA2 file
    path = Path(__file__).parents[1] / "data" / "test.WV1"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    soundfile = pytest.importorskip(  # noqa
        "soundfile", reason="If soundfile is not installed, this test will fail"
    )
    multiTrack = load_audio(path)
    wave = multiTrack["left"]
    assert np.any(wave.value > 0)
    assert wave.fs == 16000


@xfailif_no_gitlfs
@pytest.mark.xfail(reason="We cannot support this kind of file")
def test_load_wv2():
    # read WA2 file
    path = Path(__file__).parents[1] / "data" / "test.WV2"
    if is_gitlfs_pointer(path):
        pytest.skip("Audio object is a git lfs pointer")
    soundfile = pytest.importorskip(  # noqa
        "soundfile", reason="If soundfile is not installed, this test will fail"
    )
    multiTrack = load_audio(path)
    wave = multiTrack["left"]
    assert np.any(wave.value > 0)
    assert wave.fs == 16000



from .convert import ConverterToFloat64  # noqa
from .energy_estimator import EnergyEstimator  # noqa
from .f0_analyzer import F0Analyzer  # noqa
from .filter import Filter, ZeroPhaseFilter  # noqa
from .noise_reducer import NoiseReducer  # noqa
from .peak_tracker import PeakTracker  # noqa



import codecs
import logging
import os
from pathlib import Path
from typing import List, Tuple

import numpy

from signalworks.tracking import LabreadError
from signalworks.tracking.tracking import Track

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

TIME_TYPE = numpy.int64


class Partition(Track):
    default_suffix = ".lab"

    def check(self):
        assert isinstance(self._time, numpy.ndarray)
        assert self._time.ndim == 1
        assert self._time.dtype == TIME_TYPE
        # assert (numpy.diff(self._time.astype(numpy.float)) > 0).all(),\
        #     "times must be strictly monotonically increasing"
        if not (numpy.diff(self._time.astype(numpy.float)) > 0).all():
            logger.warning("Partition: times must be strictly monotonically increasing")
        assert isinstance(self._value, numpy.ndarray)
        # assert self._value.ndim == 1 # TODO: can I remove this?
        assert isinstance(self._fs, int)
        assert self._fs > 0
        # if len(self._time):
        assert self._time[0] == 0, "partition must begin at time 0"
        # assert (numpy.diff(self._time) > 0).all(), "zero-duration labels are not permitted"
        if not (numpy.diff(self._time) > 0).all():
            logger.warning("Partition: zero-duration labels are not permitted")
        # this means an empty partition contains one time value at 0!!!
        assert (
            len(self._time) == len(self._value) + 1
        ), "length of time and value+1 must match"
        # else:
        #    assert len(self._value) == 0
        return True

    def __init__(self, time, value, fs, path=None):
        super().__init__(path)
        if path is None:
            path = str(id(self))
        self.path = Path(path).with_suffix(self.default_suffix)
        self._time = time
        self._value = value
        self._fs = fs
        assert self.check()

    def get_time(self):
        # assert (numpy.diff(self._time.astype(numpy.float)) > 0).all(),\
        # "times must be strictly monotonically increasing" # in case the user messed with .time[index] directly
        if not (numpy.diff(self._time.astype(numpy.float)) > 0).all():
            logger.warning("get_time times must be strictly monotonically increasing")
        return self._time

    def set_time(self, time):
        assert isinstance(time, numpy.ndarray)
        assert time.ndim == 1
        assert time.dtype == TIME_TYPE
        if len(time):
            assert time[0] == 0, "partition must begin at time 0"
            # assert (numpy.diff(time) > 0).all(), "zero-duration labels are not permitted"
            # assert (numpy.diff(time) >= 0).all(), "zero-duration labels are not permitted"
            if not (numpy.diff(time) > 0).all():
                logger.warning("encorter zero-duration labels are not permitted")
            assert (
                len(time) == len(self._value) + 1
            ), "length of time and value+1 must match"
        else:
            assert len(self._value) == 0
        self._time = time

    time = property(get_time, set_time)

    def get_value(self):
        return self._value

    def set_value(self, value):
        assert isinstance(value, numpy.ndarray)
        # assert value.ndim == 1
        assert (
            len(self._time) == len(self._value) + 1
        ), "length of time and value must match"
        self._value = value

    value = property(get_value, set_value)

    def get_duration(self):
        if len(self._value):
            return self._time[-1]
        else:
            return 0  # or None, or raise Exception

    def set_duration(self, duration):
        assert isinstance(duration, TIME_TYPE) or isinstance(duration, int)
        if len(self._value):
            assert (
                duration > self._time[-2]
            ), "can't set duration to a smaller or equal value than the next-to-last boundary (this would result in losing the last value)"
            self._time[-1] = duration
        else:
            if duration != 0:
                raise Exception(
                    "cannot set duration of an empty Partition to anything but 0"
                )

    duration = property(get_duration, set_duration, doc="duration of track in fs")

    def get_fs(self):
        return self._fs

    def set_fs(self, _value):
        raise Exception("Cannot change fs, try resample()")

    fs = property(get_fs, set_fs, doc="sampling frequency")

    def __eq__(self, other):
        try:
            if (
                (self._fs == other._fs)
                and (len(self._time) == len(other._time))
                and (len(self._value) == len(other._value))
                and (self._time == other._time).all()
                and (self._value == other._value).all()
            ):
                return True
        except AttributeError:
            pass
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self._value)

    # def resample(self, fs):
    #     if fs != self._fs:
    #         factor = fs / self._fs
    #         self._time = numpy.round(factor * self._time).astype(TIME_TYPE)
    #         assert (numpy.diff(self._time) > 0).all(),\
    #             "new fs causes times to fold onto themselves due to lack in precision"
    #         self._fs = fs

    def resample(self, fs):
        """resample to a certain fs"""
        if fs != self._fs:
            factor = fs / self._fs
            time = numpy.round(factor * self._time).astype(TIME_TYPE)
            time[-1] = numpy.ceil(factor * self._time[-1])
            assert (
                numpy.diff(time) > 0
            ).all(), (
                "new fs causes times to fold onto themselves due to lack in precision"
            )
            return type(self)(time, self._value, fs)
        else:
            return self

    def select(self, a, b):
        assert 0 <= a
        assert a < b
        assert b <= self._time[-1]
        if __debug__:  # consistency check
            self.check()
        ai = self._time.searchsorted(a)
        bi = self._time.searchsorted(b)
        if a < self._time[ai]:  # prepend
            ai -= 1
        if self._time[bi] < b:  # append
            bi += 1
        value = self._value[ai:bi]
        time = self.time[
            ai : bi + 1
        ].copy()  # otherwise we are going to modify the original below!
        # if a < self._time[ai]:  # prepend
        time[0] = a
        # if self._time[bi] < b:  # append
        time[-1] = b
        time = time - a  # must make a copy!
        return type(self)(time.astype(TIME_TYPE), value, self.fs)

    @classmethod
    def read(cls, name, fs=300_000, track_name=None, *_args, **_kwargs):
        """Loads object from name, adding default extension if missing.
        It figures out which read function is appropriate based on extension.
        """
        if name == os.path.splitext(name)[0]:
            assert track_name, "neither track name nor extension specified"
            ext = "." + track_name
            name += ext
        else:
            ext = os.path.splitext(name)[1].lower()[1:]
        if ext in ("lab", "dem", "rec"):
            return cls.read_lab(name, fs)
        elif ext == "textgrid":
            return cls.read_textgrid(name, fs)
        else:
            try:
                return cls.read_lab(name, fs)
            except FileNotFoundError:  # TODO: need to catch more exceptions?
                pass
            try:
                return cls.read_textgrid(name, fs)
            except FileNotFoundError:  # TODO: need to catch more exceptions?
                pass
            raise ValueError("file '{}' has unknown format".format(name))

    @classmethod
    def create(
        cls, data: List[Tuple[float, float, str]], fs: int = 300_000
    ) -> "Partition":
        """
        create Partition from a list of alignment.
        The list consists of tuples in the format of (start time (sec), end time (sec), string label)
        """
        time: list = []
        value: list = []
        for (t1, t2, label) in data:
            if label[-1] == "\r":
                label = label[:-1]
            if len(time) == 0:
                time.append(t1)
            else:
                if time[-1] != t1:
                    logger.warning(f"noncontiguous label {label}")
            dur = t2 - time[-1]
            if dur > 0:
                time.append(t2)
                value.append(label)
            elif dur == 0:
                logger.warning(f"zero-length label {label}")
            else:
                raise LabreadError(
                    "label file contains times that are not monotonically increasing"
                )
        if len(time) == 0:
            raise (Exception("file is empty or all lines were ignored"))
        if time[0] != 0:
            logger.warning("moving first label boundary to zero")
            time[0] = 0
        time = numpy.round(numpy.array(time) * fs).astype(TIME_TYPE)
        value = numpy.array(
            value, dtype="U16"
        )  # if label_type is str else numpy.float64)
        return Partition(
            time, value, fs=fs
        )  # u1p to 16 characters (labels could be words)

    @classmethod
    def read_lab(
        cls, name: str, fs: int = 300_000, encoding: str = "UTF8"
    ) -> "Partition":
        """load times, values from a label file"""
        with codecs.open(name, "r", encoding=encoding) as f:
            lines = f.readlines()
        if len(lines) == 0:
            raise ValueError("file '{}' is empty".format(name))
        # label_type = numpy.float64
        data = []
        for i, line in enumerate(lines):
            try:
                tmp1, tmp2, label = line[:-1].split()
            except ValueError:
                logger.warning(
                    'ignoring line "%s" in file %s at line %i' % (line, name, i + 1)
                )
                continue
            t1 = float(tmp1)  # in second
            t2 = float(tmp2)  # in second
            data.append((t1, t2, label))
        return Partition.create(data, fs)

    @classmethod
    def read_partition(cls, name, fs=48000, encoding="UTF8"):
        """load times, values from a label file"""
        with codecs.open(name, "r", encoding=encoding) as f:
            lines = f.readlines()
        if len(lines) == 0:
            raise ValueError("file '{}' is empty".format(name))
        time = [0.0]
        value = []
        label_type = numpy.float64
        for i, line in enumerate(lines):
            try:
                temp, _t, label = line[:-1].split()
            except ValueError:
                logger.warning(
                    'ignoring line "%s" in file %s at line %i' % (line, name, i + 1)
                )
                continue
            t = float(temp)
            if label[-1] == "\r":
                label = label[:-1]
            try:
                label = label_type(label)
            except ValueError:
                if len(time) == 1:
                    label_type = str
                    label = label_type(label)
                else:
                    raise
            if len(time) == 1:
                dur = t
            else:
                dur = t - time[-1]

            time.append(t)
            if dur > 0:
                value.append(label)
            elif dur == 0:
                logger.warning(
                    'zero-length label "%s" in file %s at line %i, ignoring'
                    % (label, name, i + 1)
                )
            else:
                raise LabreadError(
                    "label file contains times that are not monotonically increasing"
                )
        if len(time) == 1:
            raise (Exception("file is empty or all lines were ignored"))
        if time[0] != 0:
            logger.warning("moving first label boundary to zero")
            time[0] = 0
            # or insert a first label
            # time.insert(0, 0)
            # value.insert(0, default_label)
        time = numpy.round(numpy.array(time) * fs).astype(TIME_TYPE)
        # assert labels are not longer than 8 characters
        value = numpy.array(value, dtype="U16" if label_type is str else numpy.float64)
        return Partition(
            time, value, fs=fs
        )  # u1p to 16 characters (labels could be words)

    def write(self, name):
        """Saves object to name, adding default extension if missing."""
        name_wo_ext = os.path.splitext(name)[0]
        if name == name_wo_ext:
            assert name, "neither track name nor extension specified"
            name += self.default_suffix
        self.write_lab(name)

    def write_lab(self, file):
        if hasattr(file, "read"):
            f = file
        else:
            f = open(file, "w")
        for i, v in enumerate(self._value):
            f.write(
                "%f %f %s\n" % (self._time[i] / self.fs, self._time[i + 1] / self.fs, v)
            )
        f.close()

    def write_partition(self, name):
        # written to generate a format exactly like
        # CMU-ARCTIC lab/*.lab files
        f = open(name, "w")
        f.write("#\n")
        for i, v in enumerate(self._value):
            f.write(f"{self._time[i + 1] / self.fs} 125 {v}\n")
        f.close()

    @classmethod
    def from_TimeValue(cls, tv):
        """convert a time value track with repeating values into a partition track"""
        # import TimeValue here to avoid circular dependencies
        from signalworks.tracking.timevalue import TimeValue

        assert isinstance(tv, TimeValue)
        boundary = numpy.where(numpy.diff(tv.value))[0]
        time = numpy.empty(len(boundary), dtype=tv.time.dtype)
        value = numpy.empty(len(boundary) + 1, dtype=tv.value.dtype)
        duration = tv.duration
        for i in range(len(time)):
            index = boundary[i]
            time[i] = (tv.time[index] + tv.time[index + 1]) / 2
        time = numpy.concatenate(([0], time, [duration])).astype(TIME_TYPE)
        value[0] = tv.value[0]
        for i in range(len(boundary)):
            value[i + 1] = tv.value[boundary[i] + 1]
        par = Partition(time, value, tv.fs, path=tv.path)
        assert par.check()
        return par

    @classmethod
    def from_Label(cls, lab, empty=''):
        """convert a Label to a partition, filling out the missing space with the empty label"""
        from signalworks.tracking.label import Label
        assert isinstance(lab, Label)
        time = []
        value = []
        # initial gap?
        if lab.time[0] != 0:
            time.append(0)
            value.append(empty)
        # first label
        time.append(lab.time[0])
        value.append(lab.value[0])
        time.append(lab.time[1])
        # medial labels
        for i in range(1, len(lab.value)):
            a = lab.time[2 * i]
            if a > time[-1]:
                value.append(empty)
                time.append(a)
            value.append(lab.value[i])
            time.append(lab.time[2 * i + 1])
        # final gap?
        if time[-1] != lab.duration:
            value.append(empty)
            time.append(lab.duration)
        par = Partition(numpy.array(time, dtype=TIME_TYPE), numpy.array(value), lab.fs, path=lab.path)
        par.check()
        return par

    def __getitem__(self, index):
        return self._time[index], self._value[index], self._time[index + 1]

    def __setitem__(self, index, value):
        raise NotImplementedError  # this would be some work to support

    def __str__(self):
        s = [""]
        for i in range(len(self._value)):
            s.append(
                "#%i: %i/%.3f: %s :%i/%.3f\n"
                % (
                    i,
                    self._time[i],
                    self._time[i] / self._fs,
                    self._value[i],
                    self._time[i + 1],
                    self._time[i + 1] / self._fs,
                )
            )
        return f"{''.join(s)}\nfs={self._fs}\nduration={self.duration}"

    def __add__(self, other):
        if self._fs != other._fs:
            raise Exception("sampling frequencies must match")
        time = numpy.hstack(
            (self._time, self._time[-1] + other._time[1:])
        )  # other._time[0] == 0
        value = numpy.hstack((self._value, other._value))
        # duration = self.duration + other.duration
        return Partition(time, value, self.fs)

    def crossfade(self, partition, length):
        """append to self, using a crossfade of a specified length in samples"""
        assert type(self) == type(
            partition
        ), "Cannot add Track objects of different types"
        assert self.fs == partition.fs, "sampling frequency of waves must match"
        assert isinstance(length, int)
        assert length > 0
        assert partition.duration >= length
        assert self.duration >= length
        # cut left, cut right, concatenate, could be sped up
        a = self.select(0, self.duration - length // 2)
        b = partition.select(length - length // 2, partition.duration)
        return a + b

    def get(self, t):
        """returns current label at time t"""
        # return self.value[numpy.where((self.time - t) <= 0)[0][-1]] # last one that is <= 0
        return self._value[
            (self._time.searchsorted(t + 1) - 1).clip(0, len(self._value) - 1)
        ]

    def append(self, time, value):
        """appends a value at the end, time is new endtime"""
        assert time > self._time[-1]
        self._time = numpy.hstack((self._time, time))
        self._value = numpy.hstack(
            (self._value, value)
        )  # so values must be numpy objects after all ?

    def insert(self, time: float, value: str) -> None:
        """modifies partition object to include value at time - other times are unchanged"""
        assert not (time == self._time).any(), "this boundary exists already"
        index = numpy.searchsorted(self._time, numpy.array([time]))[
            0
        ]  # move _this_ index to the right
        self._time = numpy.hstack((self._time[:index], time, self._time[index:]))
        # so values must be numpy objects after all ?
        self._value = numpy.hstack((self._value[:index], value, self._value[index:]))

    def delete_merge_left(self, index):
        # TODO write unittests for me and my other half /
        # or replace by just CUT if it's possible to uniquely specify the latter
        """deletes a phoneme, leaves duration as is"""
        assert len(self._value) > 1
        assert 0 <= index < len(self._value)
        self._time = numpy.hstack((self._time[:index], self._time[index + 1 :]))
        self._value = numpy.hstack((self._value[:index], self._value[index + 1 :]))

    def delete_merge_right(self, index):
        """deletes a phoneme, leaves duration as is"""
        assert len(self._value) > 1
        assert 0 <= index < len(self._value)
        self._time = numpy.hstack((self._time[: index + 1], self._time[index + 2 :]))
        self._value = numpy.hstack((self._value[:index], self._value[index + 1 :]))

    def merge_same(self):  # rename to uniq?
        """return a copy of myself, except identical values will be merged"""
        if len(self._value) <= 1:
            return Partition(self._time, self._value, self.fs)
        else:
            time = [self._time[0]]
            value = [self._value[0]]
            for i, v in enumerate(self._value[1:]):
                if v != value[-1]:  # new value
                    time.append(self._time[i + 1])  # close last one, begin new one
                    value.append(v)
            time.append(self.duration)
            return Partition(
                numpy.array(time, dtype=TIME_TYPE), numpy.array(value), self.fs
            )

    def time_warp(self, X, Y):
        assert X[0] == 0
        assert Y[0] == 0
        assert X[-1] == self.duration
        time = numpy.interp(self.time, X, Y).astype(self.time.dtype)
        # assert len(numpy.where(numpy.diff(time) <= 0)[0]) == 0,\
        #     'some segment durations are non-positive' # Must allow this after all
        # self._time = time
        # may have to remove some collapsed items
        self._time = time
        while 1:
            index = numpy.where(numpy.diff(self._time) == 0)[0]
            if len(index):
                logger.warning(
                    'removing collapsed item #{}: "{}"'.format(
                        index[0], self.value[index[0]]
                    )
                )
                self.delete_merge_right(index[0])
            else:
                break



import numpy as np
import pytest
from signalworks.tracking import Event, Track


@pytest.fixture
def var():
    t = Event(np.array([3, 6], dtype=np.int64), 1, 10)
    u = Event(np.array([0, 2], dtype=np.int64), 1, 3)
    assert isinstance(t, Event)
    assert isinstance(u, Event)
    assert isinstance(u, Track)
    return t, u


def test_init():
    with pytest.raises(Exception):
        Event(np.array([6, 3], dtype=np.int64), 1, 10)  # bad times
    with pytest.raises(Exception):
        Event(np.array([3, 6], dtype=np.int64), 1, 5)  # duration too shorts


def test_duration(var):
    t, _ = var
    t.duration = 8
    with pytest.raises(Exception):
        t.set_duration(5)


def test_eq(var):
    t, u = var
    assert t == t
    assert not t == u


def test_add(var):
    t, u = var
    v = t + u
    assert v == Event(np.array([3, 6, 10, 12], dtype=np.int64), 1, 13)
    t += u
    assert t == Event(np.array([3, 6, 10, 12], dtype=np.int64), 1, 13)


def test_select(var):
    t, u = var
    t_ = t.select(2, 7)
    assert t_ == Event(np.array([1, 4], dtype=np.int64), 1, 5)
    t_ = t.select(3, 7)
    assert t_ == Event(np.array([0, 3], dtype=np.int64), 1, 4)
    t_ = t.select(3, 6)
    assert t_ == Event(np.array([0], dtype=np.int64), 1, 3)
    t_ = t.select(2, 6)
    assert t_ == Event(np.array([1], dtype=np.int64), 1, 4)


def test_resample(benchmark):
    e = Event(np.array([3, 6], dtype=np.int64), 1, 10)
    # e = e.resample(2)
    e = benchmark(e.resample, 2)
    assert e.time[0] == 6


def test_crossfade(benchmark):
    evt1 = Event(np.array([1, 5, 9], dtype=np.int64), 1, 10)
    evt2 = Event(np.array([2, 5, 9], dtype=np.int64), 1, 10)
    length = 2
    # evt = evt1.crossfade(evt2, length)
    evt = benchmark(evt1.crossfade, evt2, length)
    assert evt1.duration + evt2.duration - length == evt.duration
    assert np.allclose(evt.time, np.array([1, 5, 10, 13, 17]))



import copy
import json
import os
from collections import UserDict

from signalworks.tracking import Event, Partition, TimeValue, Value, Wave


class MultiTrack(UserDict):
    """
    A dictionary containing time-synchronous tracks of equal duration and fs
    """

    def __init__(self, mapping=None):
        if mapping is None:
            mapping = UserDict()
        UserDict.__init__(self, mapping)
        if __debug__:  # long assert - TODO: do this on mapping, and then assign
            self.check()

    def check(self):
        if len(self) > 1:
            for i, (key, track) in enumerate(self.items()):
                if track.fs != self.fs:
                    raise AssertionError(
                        f"all fs' must be equal, track #{i} ('{key}) does not match track #1"
                    )
                if track.duration != next(iter(self.values())).duration:
                    raise AssertionError(
                        f"all durations must be equal, track #{i} ('{key}'') does not match track #1"
                    )

    def get_fs(self):
        if len(self):
            return next(iter(self.values())).fs
        else:
            return 0  # or raise?

    def set_fs(self, fs):
        raise Exception("Cannot change fs, try resample()")

    fs = property(get_fs, set_fs, doc="sampling frequency")

    def get_duration(self):
        if len(self):
            if __debug__:  # long assert - TODO: do this on mapping, and then assign
                self.check()
            return next(iter(self.values())).duration
        else:
            return 0

    def set_duration(self, duration):
        raise Exception("The duration cannot be set, it is derived from its conents")

    duration = property(
        get_duration, set_duration, doc="duration, as defined by its content"
    )

    def __eq__(self, other):
        # excluding wav from comparison as long as wav writing/reading is erroneous
        if (set(self.keys()) - {"wav"}) != (set(other.keys()) - {"wav"}):
            return False
        for k in self.keys():
            if k != "wav" and self[k] != other[k]:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __setitem__(self, key, value):
        if len(self):
            if value.duration != self.duration:
                raise AssertionError("duration does not match")
            if value.fs != self.fs:
                raise AssertionError("fs does not match")
        UserDict.__setitem__(self, key, value)

    def __str__(self):
        s = ""
        for key, track in self.items():
            s += "%s: %s\n" % (key, track)
        return s

    def __add__(self, other):
        if self is other:
            other = copy.deepcopy(other)
        obj = type(self)()
        for k in self:  # .iterkeys():
            obj[k] = self[k] + other[k]
        return obj

    def resample(self, fs):
        multiTrack = type(self)()
        for key, track in self.items():
            multiTrack[key] = track.resample(fs)
        return multiTrack

    def crossfade(self, other, length):
        """
        append multiTrack to self, using a crossfade of a specified length in samples
        """
        assert type(self) == type(other)
        assert self.keys() == other.keys()
        assert self.fs == other.fs
        assert isinstance(length, int)
        assert length > 0
        assert other.duration >= length
        assert self.duration >= length
        multiTrack = type(self)()
        for key, _ in self.items():
            multiTrack[key] = self[key].crossfade(other[key], length)
        return multiTrack

    def select(self, a, b, keys=None):
        assert a >= 0
        assert a < b  # or a <= b?
        assert b <= self.duration
        """return a new multitrack object with all track views from time a to b"""
        if keys is None:
            keys = self.keys()
        multiTrack = type(self)()
        for key in keys:
            multiTrack[key] = self[key].select(a, b)
        return multiTrack

    # TODO: should this be deprecated in favor of / should this call - the more general time_warp function?
    def scale_duration(self, factor):
        if factor != 1:
            for t in self.values():
                if isinstance(t, Partition):
                    t.time *= (
                        factor
                    )  # last time parameter IS duration, so no worries about duration
                elif isinstance(t, TimeValue) or isinstance(t, Event):
                    if factor > 1:  # make room for expanded times
                        t.duration = int(t.duration * factor)
                        t.time *= factor
                    else:
                        t.time *= factor
                        t.duration = int(t.duration * factor)
                else:
                    raise NotImplementedError  # wave?

    def time_warp(self, x, y):
        """in-place"""
        for track in iter(self.values()):
            track.time_warp(x, y)

    default_suffix = ".mtt"

    @classmethod
    def read(cls, name):
        """Loads info about stored tracks from name, adding extension if missing,
        and loads tracks by calling read(<name without extension>) for them.
        """
        name_wo_ext = os.path.splitext(name)[
            0
        ]  # TODO: upgrade all path stuff to pathlib
        if name == name_wo_ext:
            name += cls.default_suffix
        with open(name, "rb") as mtt_file:
            track_infos = json.load(mtt_file)
        self = cls()
        for track_type_name, track_info_list in track_infos:
            track_type = globals()[track_type_name]
            track_info: UserDict = UserDict(track_info_list)
            track = track_type.read(name_wo_ext, **track_info)
            self[track_info["track_name"]] = track
        return self

    @classmethod
    def read_edf(cls, path):
        raise NotImplementedError
        # TODO: adapt
        # the following is copied from elsewhere and won't work as is
        import pyedflib

        with pyedflib.EdfReader(str(path)) as f:
            labels = f.getSignalLabels()
            for label in labels:
                index = labels.index(label)
                wav = Wave(f.readSignal(index), f.getSampleFrequency(index))
                wav.label = label
                wav.path = f.with_name(f.stem + "-" + label + ".wav")
                wav.min = f.getPhysicalMinimum(index)
                wav.max = f.getPhysicalMaximum(index)
                wav.unit = f.getPhysicalDimension(index)
                # self.add_view(wav, panel_index=panel_index, y_min=wav.min, y_max=wav.max)

    @classmethod
    def read_xdf(cls, path):
        raise NotImplementedError
        import openxdf

        # TODO: below is a place holder and needs to be finalize
        xdf = openxdf.OpenXDF(path)
        signals = openxdf.Signal(xdf, path.with_suffix(".nkamp"))
        # TODO: automate this, why are the xdf.header names different from signals.list_channels?
        for label in ["ECG", "Chin"]:
            # logger.info(f'reading {label} channel')
            sig = signals.read_file(label)[label]
            wav = Wave(sig.ravel(), 200)
            wav.label = label
            # wav.path = file.with_name(file.stem + '-' + label + '.wav')
            wav.min = -3200
            wav.max = 3200
            wav.unit = "1"
            # self.add_view(wav, panel_index=panel_index, y_min=wav.min, y_max=wav.max)

    def write(self, name):
        """Saves info about stored tracks to name, adding extension if missing,
        and calls write(<name without extension>) for the contained tracks.
        Note!: not saving wav as long as wav writing/reading is erroneous
        """
        name_wo_ext = os.path.splitext(name)[0]
        if name == name_wo_ext:
            name += self.default_suffix
        track_infos = []  # list of dicts storing track info
        for track_name, track in sorted(self.items()):
            if track_name == "wav":
                continue
            track_info = {
                "track_name": track_name,
                "fs": int(track.get_fs()),
                "duration": int(track.get_duration()),
            }
            if type(track) == Value:
                track_info.update({"value_type": type(track.get_value()).__name__})
            track.write(name_wo_ext, **track_info)
            track_infos.append((type(track).__name__, sorted(track_info.items())))
        with open(name, "wt") as mtt_file:
            json.dump(track_infos, mtt_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
from itertools import chain

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

install_requires = ["numpy", "scipy", "numba"]

extras_require = {
    "test": [
        "pytest",
        "pytest-runner",
        "pytest-benchmark",
        "pytest-cov",
        "pytest-xdist",
        "coverage",
        "tox",
        "tox-conda",
        "pre-commit",
    ],
    "lint": ["mypy", "flake8", "flake8-bugbear", "flake8-comprehensions"],
    "style": ["black", "isort"],
    "docs": ["Sphinx", "watchdog"],
    "label": ["pyedflib", "openxdf"],
    "extended_audio": ["soundfile"],
}

extras_require["dev"] = list(chain.from_iterable(chain(extras_require.values())))
extras_require["full"] = list(
    chain.from_iterable(
        chain(extras_require["label"] + extras_require["extended_audio"])
    )
)

setup(
    author="Alex Kain",
    author_email="lxkain@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    description="Library to handle signal data and perform signal processing computations",
    entry_points={"console_scripts": ["signalworks=signalworks.cli:main"]},
    install_requires=install_requires,
    extras_require=extras_require,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="signalworks",
    name="signalworks",
    packages=find_packages(),
    setup_requires=["pytest-runner"],
    test_suite="tests",
    url="https://github.com/lxkain/signalworks",
    version="0.2.1",
    zip_safe=False,
)



# -*- coding: utf-8 -*-
from typing import NamedTuple, Optional, Tuple

import numpy as np
from scipy import signal
from signalworks.processors.processing import DefaultProgressTracker, Processor
from signalworks.tracking import Wave


class Filter(Processor):
    name = "Linear Filter"
    # acquire = {'wave': tracking.Wave}
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def __init__(self):
        super().__init__()
        # default is pre-emphasis
        self.parameters = {"B": np.array([1.0, -0.95]), "A": np.array([1.0])}

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[Wave]:
        # Processor.process(self, **kwargs)
        if progressTracker is not None:
            self.progressTracker = progressTracker
        wav = self.data.wave
        assert isinstance(wav, Wave)
        x = wav.value
        self.progressTracker.update(10)
        y = signal.lfilter(self.parameters["B"], self.parameters["A"], x).astype(
            x.dtype
        )
        self.progressTracker.update(90)
        new_track = (
            Wave(
                y,
                fs=wav.fs,
                path=str(
                    wav.path.with_name(wav.path.stem + "-filtered").with_suffix(
                        Wave.default_suffix
                    )
                ),
            ),
        )
        return new_track


class ZeroPhaseFilter(Filter):
    name = "Zero-phase Linear Filter"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[Wave]:
        # Processor.process(self, **kwargs)
        if progressTracker is not None:
            self.progressTracker = progressTracker
        wav = self.data.wave
        assert isinstance(wav, Wave)
        x = wav.value
        self.progressTracker.update(10)
        y = signal.filtfilt(self.parameters["B"], self.parameters["A"], x).astype(
            x.dtype
        )
        self.progressTracker.update(90)
        assert isinstance(wav, Wave)
        return (
            Wave(
                y,
                fs=wav.fs,
                path=str(
                    wav.path.with_name(wav.path.stem + "-0phasefiltered").with_suffix(
                        wav.path.suffix
                    )
                ),
            ),
        )



# -*- coding: utf-8 -*-
"""Tracks
Each track has a fs and a duration. There are 4 kinds of tracks:

1 Event - times
2 Wave - values
3 TimeValue - values at times, duration
4 Partition - values between times

All track intervals are of the type [), and duration points to the next unoccupied sample == length
"""

import logging
from builtins import str
from pathlib import Path
from typing import List, Optional, Union

import numpy
from signalworks.tracking.metatrack import MetaTrack

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
# logger.setLevel(logging.WARNING)
# logger.setLevel(logging.ERROR)

TIME_TYPE = numpy.int64


def convert_dtype(source, target_dtype):
    """
    return a link (if unchanged) or copy of signal in the specified dtype (often changes bit-depth as well)
    """
    assert isinstance(source, numpy.ndarray)
    source_dtype = source.dtype
    assert source_dtype in (
        numpy.int16,
        numpy.int32,
        numpy.float32,
        numpy.float64,
    ), "source must be a supported type"
    assert target_dtype in (
        numpy.int16,
        numpy.int32,
        numpy.float32,
        numpy.float64,
    ), "target must be a supported type"
    if source_dtype == target_dtype:
        return source
    else:  # conversion
        if source_dtype == numpy.int16:
            if target_dtype == numpy.int32:
                return source.astype(target_dtype) << 16
            else:  # target_dtype == numpy.float32 / numpy.float64:
                return source.astype(target_dtype) / (1 << 15)
        elif source_dtype == numpy.int32:
            if target_dtype == numpy.int16:
                return (source >> 16).astype(target_dtype)  # lossy
            else:  # target_dtype == numpy.float32 / numpy.float64:
                return source.astype(target_dtype) / (1 << 31)
        else:  # source_dtype == numpy.float32 / numpy.float64
            M = numpy.max(numpy.abs(source))
            limit = 1 - 1e-16
            if M > limit:
                factor = limit / M
                logger.warning(
                    f"maximum float waveform value {M} is beyond [-{limit}, {limit}],"
                    f"applying scaling of {factor}"
                )
                source *= factor
            if target_dtype == numpy.float32 or target_dtype == numpy.float64:
                return source.astype(target_dtype)
            else:
                if target_dtype == numpy.int16:
                    return (source * (1 << 15)).astype(target_dtype)  # dither?
                else:  # target_dtype == numpy.int32
                    return (source * (1 << 31)).astype(target_dtype)  # dither?


class Track(MetaTrack):
    default_suffix = ".trk"

    def __init__(self, path):
        self._fs = 0
        self.type: Optional[str] = None
        self.min: Union[int, float, None] = None
        self.max: Union[int, float, None] = None
        self.unit: Optional[str] = None
        self.label: Optional[str] = None
        if path is None:
            path = str(id(self))
        self.path = Path(path).with_suffix(self.default_suffix)

    def get_time(self):
        raise NotImplementedError

    def set_time(self, time):
        raise NotImplementedError

    time = property(get_time, set_time)

    def get_value(self):
        raise NotImplementedError

    def set_value(self, value):
        raise NotImplementedError

    value = property(get_value, set_value)

    def get_fs(self):
        return self._fs

    def set_fs(self, _value):
        raise Exception("Cannot change fs, try resample()")

    fs = property(get_fs, set_fs, doc="sampling frequency")

    def get_duration(self):
        pass

    def set_duration(self, duration):
        raise NotImplementedError

    duration = property(get_duration, set_duration)

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        raise NotImplementedError

    def __len__(self):
        pass

    def __str__(self):
        pass

    def __add__(self, other):
        raise NotImplementedError

    @classmethod
    def read(cls, path, samplerate=None):
        # we do the imports here to avoid circular import when Wave inherits Track, and Track call Wave's function
        # we only need a function from the dependencies
        from signalworks.tracking.partition import Partition
        from signalworks.tracking.timevalue import TimeValue
        from signalworks.tracking.wave import Wave
        from signalworks.tracking.multitrack import MultiTrack

        """Loads object from name, adding default extension if missing."""
        # E = []
        suffix = Path(path).suffix

        with open(path, "rb") as fileIn:
            bufHeader = fileIn.read(38)
        if (
            (bufHeader[0:4] == b"RIFF")
            and (bufHeader[12:16] == b"fmt ")
            and (bufHeader[0:5] != b"RIFFB")
        ):
            channels = None
            mmap = False
            return Wave.wav_read(path, channels, mmap)
        elif suffix == ".tmv":
            return TimeValue.read_tmv(path)  # for now, handle nans
        elif suffix == ".lab":
            return Partition.read(path)
        elif suffix == ".edf":
            return MultiTrack.read_edf(path)
        elif suffix == ".xdf":
            return MultiTrack.read_xdf(path)
        else:
            channels = None
            mmap = False
            return Wave.wav_read(path, channels, mmap)

    def write(self, name, *args, **kwargs):
        """Saves object to name, adding default extension if missing."""
        raise NotImplementedError

    def resample(self, fs):
        """resample self to a certain fs"""
        raise NotImplementedError

    def select(self, a, b):
        """
        return a selection of the track from a to b. a and b are in fs units.
        Times are new objects, but values are views - idea is to make a read-only section, not a copy
        """
        raise NotImplementedError

    def insert(self, a, t):
        raise NotImplementedError

    def remove(self, a, b):
        raise NotImplementedError

    def copy(self, a, b):
        raise NotImplementedError

    def cut(self, a, b):
        t = self.copy(a, b)
        self.remove(a, b)
        return t


def get_track_classes() -> List[Track]:
    def all_subclasses(c):
        return c.__subclasses__() + [
            a for b in c.__subclasses__() for a in all_subclasses(b)
        ]

    return [obj for obj in all_subclasses(Track)]


# TODO: class NamedEvent(_Track)
#  there hasn't been a need for it yet, but may be useful in the future
#  wonder if I can extend Event itself with optional values...
# class NamedEvent(_Track):
#  def __init__(self, time, value, fs, duration)


# class HetMultiTrack(MultiTrack):  # may want to define common abstract class instead
#     """
#     A dictionary containing time-synchronous tracks of equal duration, but HETEROGENOUS fs
#     """

#     # this fs relates to the manner by which we time-index (possibly with float) into the multitrack object.
#     # Use 1.0 for seconds.
#     def __init__(self, mapping=dict(), fs=1.0):
#         dict.__init__(self, mapping)
#         if __debug__:  # long assert - TODO: do this on mapping, and then assign
#             self.check()
#         self._fs = fs

#     def check(self):
#         if len(self) > 1:
#             duration = None
#             for i, (key, track) in enumerate(self.items()):
#                 if duration is None:
#                     duration = track.duration / track.fs
#                 if track.duration / track.fs != duration:
#                     raise AssertionError(
#                         f"all durations must be equal, track #{i} ('{key}') does not match track #1"
#                     )

#     def get_fs(self):
#         if len(self):
#             return self._fs
#         else:
#             return 0  # or raise?

#     def set_fs(self, fs):
#         self._fs = fs

#     fs = property(get_fs, set_fs, doc="sampling frequency of time-index")

#     def select(self, a, b, keys=None):
#         assert a >= 0
#         assert a < b  # or a <= b?
#         assert b <= self.duration
#         """return a new object with all track views from time a to b"""
#         if keys is None:
#             keys = self.keys()
#         obj = type(self)()
#         for key in keys:
#             trk = self[key]
#             obj[key] = trk.select(
#                 a / self._fs * trk._fs, b / self._fs * trk._fs
#             )  # untested
#         return obj

# def test_pml(self):
#     import tempfile
#     tmp = tempfile.NamedTemporaryFile(prefix='test_pml_')
#     filename = tmp.name
#     tmp.close()
#     self.t.pmlwrite(filename)
#     s = Event.pmlread(filename)
#     os.unlink(filename)
#     # duration CANNOT be encoded in the file (or can it?)
#     s.duration = int(numpy.round(self.t.duration * s.fs / self.t.fs))
#     s = s.resample(self.t.fs)
#     self.assertTrue(numpy.allclose(s.time, self.t.time))



# -*- coding: utf-8 -*-
import numpy as np
import pytest  # noqa: F401

from signalworks import dsp
from signalworks.tracking.wave import Wave


def test_spectrogram(speech_track, benchmark):
    if speech_track is None:
        pytest.skip("the test need wavefile")
    frame_size = 0.030
    half = frame_size * speech_track.fs // 2
    centers = np.round(np.linspace(half, speech_track.duration - half, 1000)).astype(
        np.int
    )

    NFFT = 2 ** dsp.nextpow2(512 * 2)
    assert isinstance(speech_track, Wave)
    X, f = benchmark(
        dsp.spectrogram_centered,
        speech_track,
        frame_size,
        centers,
        NFFT=NFFT,
        normalized=False,
    )
    assert X.shape == (len(centers), NFFT // 2 + 1)


#
#
# def test_world(speech_track, benchmark):
#     sp, f0, t = benchmark(dsp.world, speech_track)
#     assert sp.shape[1] == 513



# -*- coding: utf-8 -*-
import logging
from pathlib import Path
from typing import List, Optional

import numpy as np
from scipy.io.wavfile import read as wav_read

from .error import LabreadError, MultiChannelError  # noqa
from .event import Event  # noqa
from .label import Label  # noqa
from .metatrack import MetaTrack  # noqa
from .partition import Partition  # noqa
from .timevalue import TimeValue  # noqa
from .tracking import Track  # noqa: F401
from .value import Value  # noqa
from .wave import Wave  # noqa

from .multitrack import MultiTrack  # isort:skip

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


def load_audio(
    path: Path,
    channel: Optional[int] = None,
    mmap: bool = False,
    channel_names: List[str] = ["left", "right"],
) -> MultiTrack:
    """load waveform from file"""
    multiTrack = MultiTrack()
    assert 0 < len(channel_names) <= 2
    try:
        fs, value = wav_read(path, mmap=mmap)
    except ValueError:
        try:
            import soundfile as sf

            value, fs = sf.read(path, dtype="int16")
        except ImportError:
            logging.error(
                f"Scipy was unable to import {path}, "
                f"try installing soundfile python package for more compatability"
            )
            raise ImportError
        except RuntimeError:
            raise RuntimeError(f"Unable to import audio file {path}")
    if value.ndim == 1:
        if channel is not None and channel != 0:
            raise MultiChannelError(
                f"cannot select channel {channel} from monaural file {path}"
            )
        multiTrack[channel_names[0]] = Wave(value, fs, path=str(path))
    if value.ndim == 2:

        if channel is None:
            multiTrack[channel_names[0]] = Wave(value[:, 0], fs, path=str(path))
            multiTrack[channel_names[1]] = Wave(value[:, 1], fs, path=str(path))
        else:
            try:
                multiTrack[channel_names[channel]] = Wave(
                    value[:, channel], fs, path=str(path)
                )
            except IndexError:
                raise MultiChannelError(
                    f"cannot select channel {channel} from file "
                    f"{path} with {value.shape[1]} channels"
                )

    for k in multiTrack.keys():
        value = multiTrack[k].value

        if value.dtype == np.dtype(np.int16):
            multiTrack[k].min = -32767
            multiTrack[k].max = 32768
        elif value.dtype == np.dtype(np.int32):
            multiTrack[k].min = -2147483648
            multiTrack[k].max = 2147483647
        elif value.dtype == np.dtype(np.uint8):
            multiTrack[k].min = 0
            multiTrack[k].max = 255
        elif value.dtype in set([np.dtype(np.float64), np.dtype(np.float32)]):
            multiTrack[k].max = 1.0
            multiTrack[k].min = -1.0
        else:
            logging.error(f"Wave dtype {value.dtype} not supported")
            raise NotImplementedError
    return multiTrack
